/*
 * Copyright 1993-2006 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:   
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and 
 * international Copyright laws.  
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE 
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR 
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH 
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF 
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.   
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL, 
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS 
 * OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE 
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE 
 * OR PERFORMANCE OF THIS SOURCE CODE.  
 *
 * U.S. Government End Users.  This source code is a "commercial item" as 
 * that term is defined at 48 C.F.R. 2.101 (OCT 1995), consisting  of 
 * "commercial computer software" and "commercial computer software 
 * documentation" as such terms are used in 48 C.F.R. 12.212 (SEPT 1995) 
 * and is provided to the U.S. Government only as a commercial end item.  
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through 
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the 
 * source code with only those rights set forth herein.
 */

#ifndef _BOXFILTER_KERNEL_H_
#define _BOXFILTER_KERNEL_H_

texture<float, 2> tex;

/*
    Perform a fast box filter using the sliding window method.

    As the kernel moves from left to right, we add in the contribution of the new
    sample on the right, and subtract the value of the exiting sample on the left.
    This only requires 2 adds and a mul per output value, independent of the filter radius.
    The box filter is separable, so to perform a 2D box filter we perform the filter in
    the x direction, followed by the same filter in the y direction.
    Applying multiple iterations of the box filter converges towards a Gaussian blur.
    Using CUDA, rows or columns of the image are processed in parallel.
    This version duplicates edge pixels.

    Note that the x (row) pass suffers from uncoalesced global memory reads,
    since each thread is reading from a different row. For this reason it is
    better to use texture lookups for the x pass.
    The y (column) pass is perfectly coalesced.

    Parameters
    id - pointer to input data in global memory
    od - pointer to output data in global memory
    w  - image width
    h  - image height
    r  - filter radius

    e.g. for r = 2, w = 8:

    0 1 2 3 4 5 6 7
    x - -
    - x - -
    - - x - -
      - - x - -
        - - x - -
          - - x - -
            - - x -
              - - x
*/

__device__ void
d_boxfilter_x(float *id, float *od, int w, int h, int r)
{
    float scale = 1.0f / (2*r+1);

    float t;
    // do left edge
    t = id[0] * r;
    for (int x = 0; x < r+1; x++) {
        t += id[x];
    }
    od[0] = t * scale;

    for(int x = 1; x < r+1; x++) {
        t += id[x+r];
        t -= id[0];
        od[x] = t * scale;
    }
    
    // main loop
    for(int x = r+1; x < w-r; x++) {
        t += id[x+r];
        t -= id[x-r-1];
        od[x] = t * scale;
    }

    // do right edge
    for (int x = w-r; x < w; x++) {
        t += id[w-1];
        t -= id[x-r-1];
        od[x] = t * scale;
    }
}

__device__ void
d_boxfilter_y(float *id, float *od, int w, int h, int r)
{
    float scale = 1.0f / (2*r+1);
    
    float t;
    // do left edge
    t = id[0] * r;
    for (int y = 0; y < r+1; y++) {
        t += id[y*w];
    }
    od[0] = t * scale;

    for(int y = 1; y < r+1; y++) {
        t += id[(y+r)*w];
        t -= id[0];
        od[y*w] = t * scale;
    }
    
    // main loop
    for(int y = r+1; y < h-r; y++) {
        t += id[(y+r)*w];
        t -= id[((y-r)*w)-w];
        od[y*w] = t * scale;
    }

    // do right edge
    for (int y = h-r; y < h; y++) {
        t += id[(h-1)*w];
        t -= id[((y-r)*w)-w];
        od[y*w] = t * scale;
    }
}

__global__ void
d_boxfilter_x_global(float *id, float *od, int w, int h, int r)
{
	unsigned int y = blockIdx.x*blockDim.x + threadIdx.x;
    d_boxfilter_x(&id[y*w], &od[y*w], w, h, r);
}

__global__ void
d_boxfilter_y_global(float *id, float *od, int w, int h, int r)
{
	unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
	d_boxfilter_y(&id[x], &od[x], w, h, r);
}

// texture version
// texture fetches automatically clamp to edge of image
__global__ void
d_boxfilter_x_tex(float *od, int w, int h, int r)
{
    float scale = 1.0f / (2*r+1);
	unsigned int y = blockIdx.x*blockDim.x + threadIdx.x;
    float t = 0.0f;
    for(int x=-r; x<=r; x++) {
        t += tex2D(tex, x, y);
    }
    od[y*w] = t * scale;
    for(int x=1; x<w; x++) {
        t += tex2D(tex, x + r, y);
        t -= tex2D(tex, x - r - 1, y);
        od[y*w+x] = t * scale;
    }
}

__global__ void
d_boxfilter_y_tex(float *od, int w, int h, int r)
{
    float scale = 1.0f / (2*r+1);
	unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
    float t = 0.0f;
    for(int y=-r; y<=r; y++) {
        t += tex2D(tex, x, y);
    }
    od[x] = t * scale;
    for(int y=1; y<h; y++) {
        t += tex2D(tex, x, y + r);
        t -= tex2D(tex, x, y - r - 1);
        od[y*w+x] = t * scale;
    }
}

#endif // #ifndef _BOXFILTER_KERNEL_H_
